'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _xmlParser = require('xml-parser');

var _xmlParser2 = _interopRequireDefault(_xmlParser);

var _Text = require('./Text');

var _Text2 = _interopRequireDefault(_Text);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// SimpleNodeObj and SimpleNode are simple ways of representing a Node-like
// or Tree-like data-structure that can be easily transformed into a full `Node`


// @see: SimpleNodeObj


// Options type for Node
const xmlToSimpleNode = root => {
  return {
    value: root.attributes.value || root.content || root.name,
    children: root.children.map(xmlToSimpleNode)
  };
};

class Node {

  /**
   * Create a `Node` from a `SimpleNode`. Applied recursively to children.
   */
  static fromSimpleNode(simpleNode, parent = null, position = 0) {
    const options = {
      position,
      parent,
      value: typeof simpleNode === 'string' ? simpleNode : simpleNode.value
    };

    const node = new Node(options);

    if (typeof simpleNode !== 'string' && simpleNode.children) {
      node.children = simpleNode.children.map((child, i) => Node.fromSimpleNode(child, node, i));
    }

    return node;
  }

  static fromXML(xml) {
    const xmlRoot = (0, _xmlParser2.default)(xml);
    const simpleRoot = xmlToSimpleNode(xmlRoot.root);
    return Node.fromSimpleNode(simpleRoot);
  }

  /**
   * Supply optional values based on the public properties in `Node`.
   */
  constructor(node) {
    this.position = 0;
    this.parent = null;
    this.value = '';
    this.children = [];
    this.width = 0;
    this.left = 0;

    Object.assign(this, node);

    this.text = new _Text2.default(this.value);
  }

  /**
   * Get the previous sibling of the current `Node`, if exists,
   * otherwise `null`.
   */
  prevSib() {
    if (this.parent && this.position > 0) {
      return this.parent.children[this.position - 1];
    }
    return null;
  }

  /**
   * Get a level representation of the `Node` and its children. The levels
   * are represented as a matrix of `Node`s with the first level containing
   * a simple element, the root. Each subsequent level contains the children
   * from _ALL_ of the nodes in the preceding level.
   */
  getLevels() {
    let queue = [this];
    const ret = [];
    while (queue.length) {
      ret.push(queue);
      const temp = [];
      queue.forEach(n => {
        temp.splice(temp.length, n.children.length, ...n.children);
      });
      queue = temp;
    }
    return ret;
  }

  /**
   * Iterate through all of the children of a given `Node`
   */
  forEachChild(visitor) {
    const queue = [...this.children];
    while (queue.length) {
      const node = queue.shift();
      visitor(node);
      queue.splice(queue.length, node.children.length, ...node.children);
    }
  }
}
exports.default = Node;