#!/usr/bin/env node
'use strict';

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _yargs = require('yargs');

var _yargs2 = _interopRequireDefault(_yargs);

var _package = require('../package.json');

var _package2 = _interopRequireDefault(_package);

var _Tree = require('./Tree');

var _Tree2 = _interopRequireDefault(_Tree);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const EXEC_NAME = 'ascii-tree';

const VERSION = _package2.default.version;

const helpMsg = `
${EXEC_NAME} ${VERSION}
Usage:
  ${EXEC_NAME} [...options] <file>

Options:
  --help           Display this message.
  --space          The number of spaces between nodes
  --vertical       The character to use for drawing vertical branches
  --horizontal     The character to use for drasing horizontal branches
  --verticalHeight The number of vertical characters to draw for a single vertical branch
  --align          The text alignment (center, left)
`;

/**
 * Display help message to stderr and exit with error code 1.
 */
const help = () => {
  console.error(helpMsg);
  process.exit(1);
};

/**
 * Promisify `fs.readFile`. Reads as UTF-8.
 * @param  {String} fileName The file.
 * @return {Promise<string>}
 */
const readFile = fileName => new Promise((resolve, reject) => {
  _fs2.default.readFile(fileName, 'utf-8', (err, data) => {
    if (err) return reject(err);
    resolve(data);
  });
});

/**
 * Parse the options from `yargs.argv`.
 * @type {Promise<Options>}
 */
const getOptions = async () => {
  const { argv } = _yargs2.default;
  const fileName = argv._[0];
  const file = await readFile(fileName);

  let options;

  if (_path2.default.extname(fileName) === '.xml') {
    options = { root: file };
  } else {
    try {
      options = { root: JSON.parse(file) };
    } catch (err) {
      throw new Error(`Bad File "${fileName}". Please provide xml or JSON formats.`);
    }
  }

  if (argv.space) {
    options.space = parseInt(argv.space, 10);
    if (options.space < 0) {
      throw new Error(`Bad Option (space) "${options.space}": ` + `Please provide a non-negative value.`);
    }
  }
  if (argv.verticalHeight) {
    options.verticalHeight = parseInt(argv.verticalHeight, 10);
    if (options.verticalHeight < 0) {
      throw new Error(`Bad Option (verticalHeight) "${options.verticalHeight}": ` + `Please provide a non-negative value.`);
    }
  }
  if (typeof argv.vertical === 'string') {
    options.vertical = argv.vertical;
  }
  if (typeof argv.horizontal === 'string') {
    options.horizontal = argv.horizontal;
  }
  if (typeof argv.align === 'string') {
    const { align } = argv;
    if (align !== 'left' && align !== 'center') {
      throw new Error(`Bad Option (align) "${align}": ` + `Possible values are "center", or "left".`);
    }
    options.align = align;
  }

  return options;
};

/**
 * Main CLI entry.
 */
const main = async () => {
  const { argv } = _yargs2.default;
  if (argv.h || argv.help) {
    return help();
  }

  try {
    const options = await getOptions();
    const tree = new _Tree2.default(options);
    console.log(tree.render());
  } catch (err) {
    console.error(err.message);
    process.exit(1);
  }
};

// Run main, if error, throw on next tick.
main().catch(err => {
  process.nextTick(() => {
    throw err;
  });
});